<?php
namespace nur\sery\wip\schema;

use ArrayAccess;
use IteratorAggregate;
use nulib\php\func;
use nur\sery\wip\schema\_assoc\AssocWrapper;
use nur\sery\wip\schema\_list\ListWrapper;
use nur\sery\wip\schema\_scalar\ScalarResult;
use nur\sery\wip\schema\_scalar\ScalarWrapper;
use nur\sery\wip\schema\input\Input;
use nur\sery\wip\schema\types\IType;

abstract class Wrapper implements ArrayAccess, IteratorAggregate {
  function isAssoc(?AssocWrapper &$wrapper=null): bool { return false; }
  function isList(?ListWrapper &$wrapper=null): bool { return false; }
  function isScalar(?ScalarWrapper &$wrapper=null): bool { return false; }

  protected WrapperContext $context;

  /** changer les paramètres de gestion des valeurs */
  function resetParams(?array $params): void {
    $this->context->resetParams($params);
  }

  protected function afterModify(?array $params, bool $reset=true): void {
    $context = $this->context;
    if ($reset) {
      $context->type = null;
      $context->result->reset();
      $context->analyzed = false;
      $context->normalized = false;
    }
    if ($params["analyze"] ?? $context->analyze) {
      $this->analyze($params);
    }
    if ($context->analyzed) {
      if ($params["normalize"] ?? $context->normalize) {
        $this->normalize($params);
      }
    }
  }

  protected function newInput(&$value): Input {
    return new Input($value);
  }

  /**
   * spécifier la valeur destination gérée par cet objet.
   *
   * @param ?array $params paramètres spécifique à cet appel, qui peuvent être
   * différent des paramètres par défaut
   */
  abstract function reset(&$value, $valueKey=null, ?array $params=null): self;

  /** analyser la valeur */
  abstract protected function _analyze(?array $params): int;

  function analyze(?array $params=null): bool {
    $context = $this->context;
    $reanalyze = $params["reanalyze"] ?? false;
    if ($context->analyzed && !$reanalyze) return false;

    $this->_analyze($params);
    $context->analyzed = true;
    return true;
  }

  /** normaliser la valeur */
  abstract protected function _normalize(?array $params): bool;

  function normalize(?array $params=null): bool {
    $context = $this->context;

    // il faut que la valeur soit analysée avant de la normaliser
    $this->analyze($params);
    if (!$context->analyzed) return false;

    $renormalize = $params["renormalize"] ?? false;
    if ($renormalize || !$context->normalized) {
      $modified = $this->_normalize($params);
      $context->normalized = true;
    } else {
      $modified = false;
    }

    /** @var ScalarResult $result */
    $result = $context->result;
    if (!$result->valid) {
      $result->throw($params["throw"] ?? $context->throw);
    }
    return $modified;
  }

  /**
   * Obtenir la liste des clés valides pour les valeurs accessibles via cet
   * objet
   */
  abstract function getKeys(): array;

  /**
   * sélectionner le wrapper associé à la clé spécifiée
   *
   * @param string|int|null $key
   * @return Wrapper $this
   */
  abstract function select($key): Wrapper;

  function getIterator() {
    foreach ($this->getKeys() as $key) {
      yield $key => $this->select($key);
    }
    $this->select(null);
  }

  /**
   * obtenir le résultat de l'analyse de la valeur du wrapper sélectionné
   *
   * cette fonction doit être appelée après {@link set()} ou {@link unset()} et
   * après que le wrapper aie été sélectionné avec {@link select()}
   */
  function getResult(): Result {
    return $this->context->result;
  }

  /** retourner true si la valeur existe */
  function isPresent(): bool {
    return $this->context->result->present;
  }

  /** retourner le type associé à la valeur */
  function getType(): IType {
    return $this->context->type;
  }

  /** retourner true si la valeur est disponible */
  function isAvailable(): bool {
    return $this->context->result->available;
  }

  /** retourner true si la valeur est valide */
  function isValid(): bool {
    return $this->context->result->valid;
  }

  /** retourner true si la valeur est dans sa forme normalisée */
  function isNormalized(): bool {
    return $this->context->result->normalized;
  }


  function get($default=null) {
    $context = $this->context;
    if ($context->result->available) return $context->input->get($context->valueKey);
    else return $default;
  }

  function set($value, ?array $params=null): self {
    $context = $this->context;
    $context->input->set($value, $context->valueKey);
    $this->afterModify($params);
    return $this;
  }

  function unset(?array $params=null): self {
    $context = $this->context;
    $context->input->unset($context->valueKey);
    $this->afterModify($params);
    return $this;
  }

  protected function _format(WrapperContext $context, $format=null): string {
    $value = $context->input->get($context->valueKey);
    /** @var func $formatterFunc */
    $formatterFunc = $context->schema->formatterFunc;
    if ($formatterFunc !== null) {
      # la fonction formatter n'a pas forcément accès au format de la définition
      # le lui fournir ici
      $format ??= $context->schema->format;
      return $formatterFunc->invoke([$value, $format, $context, $this]);
    } else {
      # on assume que le type a été initialisé avec le format de la définition
      # le cas échéant
      return $context->type->format($value, $format);
    }
  }

  /** formatter la valeur pour affichage */
  function format($format=null): string {
    return $this->_format($this->context, $format);
  }

  #############################################################################
  # key & properties

  function offsetExists($offset): bool {
    return in_array($offset, $this->getKeys());
  }

  function offsetGet($offset) {
    return $this->select($offset)->get();
  }

  function offsetSet($offset, $value): void {
    $this->select($offset)->set($value);
  }

  function offsetUnset($offset): void {
    $this->select($offset)->unset();
  }
}
